#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Random;

namespace HeuristicLab.Algorithms.DataAnalysis {
  /// <summary>
  /// Linear regression data analysis algorithm.
  /// </summary>
  [Item("TSNE", "t-distributed stochastic neighbourhood embedding projects the data in a low dimensional space to allow visual cluster identification")]
  [Creatable(CreatableAttribute.Categories.DataAnalysis, Priority = 100)]
  [StorableClass]
  public sealed class TSNEAnalysis : FixedDataAnalysisAlgorithm<IRegressionProblem> {

    #region Resultnames
    private const string ScatterPlotResultName = "Scatterplot";
    private const string DataResultName = "Projected Data";
    #endregion

    #region Parameternames
    private const string DistanceParameterName = "DistanceFunction";
    private const string PerplexityParameterName = "Perplexity";
    private const string ThetaParameterName = "Theta";
    private const string NewDimensionsParameterName = "Dimensions";
    private const string MaxIterationsParameterName = "MaxIterations";
    private const string StopLyingIterationParameterName = "StopLyingIteration";
    private const string MomentumSwitchIterationParameterName = "MomentumSwitchIteration";
    private const string InitialMomentumParameterName = "InitialMomentum";
    private const string FinalMomentumParameterName = "FinalMomentum";
    private const string EtaParameterName = "Eta";
    private const string SetSeedRandomlyParameterName = "SetSeedRandomly";
    private const string SeedParameterName = "Seed";
    #endregion

    #region Parameterproperties
    public IFixedValueParameter<DoubleValue> PerplexityParameter
    {
      get { return Parameters[PerplexityParameterName] as IFixedValueParameter<DoubleValue>; }
    }
    public IFixedValueParameter<DoubleValue> ThetaParameter
    {
      get { return Parameters[ThetaParameterName] as IFixedValueParameter<DoubleValue>; }
    }
    public IFixedValueParameter<IntValue> NewDimensionsParameter
    {
      get { return Parameters[NewDimensionsParameterName] as IFixedValueParameter<IntValue>; }
    }
    public IValueParameter<IDistance<RealVector>> DistanceParameter
    {
      get { return Parameters[DistanceParameterName] as IValueParameter<IDistance<RealVector>>; }
    }
    public IFixedValueParameter<IntValue> MaxIterationsParameter
    {
      get { return Parameters[MaxIterationsParameterName] as IFixedValueParameter<IntValue>; }
    }
    public IFixedValueParameter<IntValue> StopLyingIterationParameter
    {
      get { return Parameters[StopLyingIterationParameterName] as IFixedValueParameter<IntValue>; }
    }
    public IFixedValueParameter<IntValue> MomentumSwitchIterationParameter
    {
      get { return Parameters[MomentumSwitchIterationParameterName] as IFixedValueParameter<IntValue>; }
    }
    public IFixedValueParameter<DoubleValue> InitialMomentumParameter
    {
      get { return Parameters[InitialMomentumParameterName] as IFixedValueParameter<DoubleValue>; }
    }
    public IFixedValueParameter<DoubleValue> FinalMomentumParameter
    {
      get { return Parameters[FinalMomentumParameterName] as IFixedValueParameter<DoubleValue>; }
    }
    public IFixedValueParameter<DoubleValue> EtaParameter
    {
      get { return Parameters[EtaParameterName] as IFixedValueParameter<DoubleValue>; }
    }
    public IFixedValueParameter<BoolValue> SetSeedRandomlyParameter
    {
      get { return Parameters[SetSeedRandomlyParameterName] as IFixedValueParameter<BoolValue>; }
    }
    public IFixedValueParameter<IntValue> SeedParameter
    {
      get { return Parameters[SeedParameterName] as IFixedValueParameter<IntValue>; }
    }
    #endregion

    #region  Properties
    public IDistance<RealVector> Distance
    {
      get { return DistanceParameter.Value; }
    }
    public double Perplexity
    {
      get { return PerplexityParameter.Value.Value; }
    }
    public double Theta
    {
      get { return ThetaParameter.Value.Value; }
    }
    public int NewDimensions
    {
      get { return NewDimensionsParameter.Value.Value; }
    }
    public int MaxIterations
    {
      get { return MaxIterationsParameter.Value.Value; }
    }
    public int StopLyingIteration
    {
      get { return StopLyingIterationParameter.Value.Value; }
    }
    public int MomentumSwitchIteration
    {
      get { return MomentumSwitchIterationParameter.Value.Value; }
    }
    public double InitialMomentum
    {
      get { return InitialMomentumParameter.Value.Value; }
    }
    public double FinalMomentum
    {
      get { return FinalMomentumParameter.Value.Value; }
    }
    public double Eta
    {
      get { return EtaParameter.Value.Value; }
    }
    public bool SetSeedRandomly
    {
      get { return SetSeedRandomlyParameter.Value.Value; }
    }
    public uint Seed
    {
      get { return (uint)SeedParameter.Value.Value; }
    }
    #endregion

    #region Constructors & Cloning
    [StorableConstructor]
    private TSNEAnalysis(bool deserializing) : base(deserializing) { }
    private TSNEAnalysis(TSNEAnalysis original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) { return new TSNEAnalysis(this, cloner); }
    public TSNEAnalysis() {
      Problem = new RegressionProblem();
      Parameters.Add(new ValueParameter<IDistance<RealVector>>(DistanceParameterName, "The distance function used to differentiate similar from non-similar points", new EuclidianDistance()));
      Parameters.Add(new FixedValueParameter<DoubleValue>(PerplexityParameterName, "Perplexity-Parameter of TSNE. Comparable to k in a k-nearest neighbour algorithm", new DoubleValue(25)));
      Parameters.Add(new FixedValueParameter<DoubleValue>(ThetaParameterName, "Value describing how much appoximated gradients my differ from exact gradients. Set to 0 for exact calculation", new DoubleValue(0.1)));
      Parameters.Add(new FixedValueParameter<IntValue>(NewDimensionsParameterName, "Dimensionality of projected space (usually 2 for easy visual analysis", new IntValue(2)));
      Parameters.Add(new FixedValueParameter<IntValue>(MaxIterationsParameterName, "Maximum number of iterations for gradient descent", new IntValue(1000)));
      Parameters.Add(new FixedValueParameter<IntValue>(StopLyingIterationParameterName, "Number of iterations after which p is no longer approximated", new IntValue(250)));
      Parameters.Add(new FixedValueParameter<IntValue>(MomentumSwitchIterationParameterName, "Number of iterations after which the momentum in the gradient descent is switched", new IntValue(250)));
      Parameters.Add(new FixedValueParameter<DoubleValue>(InitialMomentumParameterName, "The initial momentum in the gradient descent", new DoubleValue(0.5)));
      Parameters.Add(new FixedValueParameter<DoubleValue>(FinalMomentumParameterName, "The final momentum", new DoubleValue(0.8)));
      Parameters.Add(new FixedValueParameter<DoubleValue>(EtaParameterName, "Gradient Descent learning rate", new DoubleValue(200)));
      Parameters.Add(new FixedValueParameter<BoolValue>(SetSeedRandomlyParameterName, "If the seed should be random", new BoolValue(true)));
      Parameters.Add(new FixedValueParameter<IntValue>(SeedParameterName, "The seed used if it should not be random", new IntValue(0)));
    }
    #endregion

    protected override void Run() {
      var lowDimData = new DoubleMatrix(GetProjectedData(Problem.ProblemData));
      Results.Add(new Result(ScatterPlotResultName, "Plot of the projected data", CreateScatterPlot(lowDimData, Problem.ProblemData)));
      Results.Add(new Result(DataResultName, "Projected Data", lowDimData));
    }

    private ScatterPlot CreateScatterPlot(DoubleMatrix lowDimData, IDataAnalysisProblemData problemData) {
      var plot = new ScatterPlot(DataResultName, "");
      Normalize(lowDimData);
      plot.Rows.Add(new ScatterPlotDataRow("Training", "Points of the training set", problemData.TrainingIndices.Select(i => new Point2D<double>(lowDimData[i, 0], lowDimData[i, 1]))));
      plot.Rows.Add(new ScatterPlotDataRow("Test", "Points of the test set", problemData.TestIndices.Select(i => new Point2D<double>(lowDimData[i, 0], lowDimData[i, 1]))));
      return plot;
    }

    private double[,] GetProjectedData(IDataAnalysisProblemData problemData) {
      var random = SetSeedRandomly ? new MersenneTwister() : new MersenneTwister(Seed);
      var tsne = new TSNE<RealVector>(Distance, random, Results, MaxIterations, StopLyingIteration, MomentumSwitchIteration, InitialMomentum, FinalMomentum, Eta);
      var dataset = problemData.Dataset;
      var allowedInputVariables = problemData.AllowedInputVariables.ToArray();
      var data = new RealVector[dataset.Rows];
      for (var row = 0; row < dataset.Rows; row++) data[row] = new RealVector(allowedInputVariables.Select(col => dataset.GetDoubleValue(col, row)).ToArray());
      return tsne.Run(data, NewDimensions, Perplexity, Theta);
    }

    private static void Normalize(DoubleMatrix data) {
      var max = new double[data.Columns];
      var min = new double[data.Columns];
      for (var i = 0; i < max.Length; i++) max[i] = min[i] = data[0, i];
      for (var i = 0; i < data.Rows; i++)
        for (var j = 0; j < data.Columns; j++) {
          var v = data[i, j];
          max[j] = Math.Max(max[j], v);
          min[j] = Math.Min(min[j], v);
        }
      for (var i = 0; i < data.Rows; i++) {
        for (var j = 0; j < data.Columns; j++) {
          data[i, j] = (data[i, j] - (max[j] + min[j]) / 2) / (max[j] - min[j]);
        }
      }

    }
  }
}
